/*
 * This is SSSim: the Simple & Stupid Simulator
 *
 *  Copyright (c) 2008 Luca Abeni
 *  Copyright (c) 2008 Csaba Kiraly
 *
 *  This is free software; see gpl-3.0.txt
 */

#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "config.h"
#include "core.h"
#include "graph.h"
#include "sched.h"
#include "stats.h"
#include "td.h"

static struct peer *peers;
extern FILE *resfile;
extern FILE *statusfile;
static FILE *statsfile;
static FILE *graphfile;

int PO_DELAY;    // playout delay
static int num_peers = N;   //number of nodes
static int num_chunks = K;   //number of chunks
static int nn;
static unsigned int seed = 1;
static int buf_size;
static int ts = 2;
static int display_neighbourhood;

static void show_usage(char *argv[])
{
  printf("Usage: %s [options]\n", argv[0]);

  printf("Options:\n");
  printf("\t-s <seed>:\t\t\tSet the seed for the random numbers generator\n");
  printf("\t-c <number of chunks>:\t\tSet the number of chunks\n");
  printf("\t-o <neighbourhood size>:\tSet the neighbourhood size (outdegree)\n");
  printf("\t-n <number of peers>:\t\tSet the number of peers\n");
  printf("\t-t <trace file>\n");
  printf("\t-r <results file>:\t\t(dynamically generated during simulation)\n");
  printf("\t-p <status file>\n");
  printf("\t-P:\t\t\tPrint the neighbourhoods\n");
  printf("\t-S <stats file>\n");
  printf("\t-b <chunks buffer size> (playout delay)\n");
  printf("\t-T <Server Period>:\t\tSet the deadline postponing amount\n");
  printf("\t-g <Graph file name>:\t\tSave the overlay as a DOT file\n");

  exit(0);
}

static int opts_parse(int argc, char *argv[])
{
  int o;
  
  while ((o = getopt(argc, argv, "c:o:n:t:r:s:p:S:b:T:g:hP")) != -1) {
    switch(o) {
      case 'n':
        num_peers = atoi(optarg);
        break;
      case 'c':
        num_chunks = atoi(optarg);
        break;
      case 'o':
        nn = atoi(optarg);
        break;
      case 'b':
        buf_size = pow(2, ilogb(2 * atoi(optarg) - 1));
        break;
      case 't':
        if (strcmp(optarg, "-") == 0) {
          trace = stdout;
        } else {
          trace = fopen(optarg, "w");
        }
        break;
      case 'r':
        resfile = fopen(optarg, "w");
        break;
      case 'p':
        statusfile = fopen(optarg, "w");
        break;
      case 'P':
        display_neighbourhood = 1;
        break;
      case 'S':
        statsfile = fopen(optarg, "w");
        break;
      case 's':
        seed = atoi(optarg);
        if (seed == 0) {
          seed = time(NULL);
        }
        break;
      case 'T':
        ts = atoi(optarg);
        if (ts <= 1) {
          fprintf(stderr, "Invalid Server Period %d\n", ts);
          exit(-1);
        }
        break;
      case 'g':
        graphfile = fopen(optarg, "w");
        break;
      case 'h':
        show_usage(argv);
      default:
	fprintf(stderr, "Error: unknown option %c\n", o);

	exit(-1);
      }
  }
  if (nn == 0) {
    nn = num_peers - 1;
  }

  return optind - 1;
}




///==========================================================
///     main
///==========================================================

int main(int argc, char *argv[])
{
  int c, n;
  struct chunk *ch;
  struct peer source;

  n = opts_parse(argc, argv);
  argc -= n;
  argv += n;

  srand(seed);
 
  printf("\nN=%d\n", num_peers);
  if (trace) fprintf(trace,"\n N=%d\n", num_peers);
  peers = malloc(num_peers * sizeof(struct peer));
  if (peers == NULL) {
    perror("MAlloc");

    return -1;
  }
  memset(peers, num_peers * sizeof(struct peer), 0);

    PO_DELAY = ilogb(2*num_peers-1)+1;
#ifdef NO_BUFFER
    buf_size = num_chunks;
#else
    if (buf_size == 0) {
      buf_size = pow(2, 1 + ilogb(2*PO_DELAY-1));
    }
#endif
    for (n = 0; n < num_peers; n++) {
      peers[n].chunks = malloc(buf_size * sizeof(struct chunk_instance *));
      peers[n].recv_chunks = malloc(buf_size * sizeof(struct chunk_instance *));
      peers[n].num_chunks = num_chunks;
      peers[n].buf_size = buf_size;
      peers[n].buf_mask = peers[n].buf_size - 1;
      for (c = 0; c < /*num_chunks*/ buf_size; c++) {
        peers[n].chunks[c] = NULL;
        peers[n].recv_chunks[c] = NULL;
        peers[n].latest_not_needed = -1; //CACHE latest_not_needed
      }
      peers[n].id = n;
      peers[n].neighbour = neighbourhood_generate(nn, peers, num_peers, n);
      peers[n].neigh_size = nn;
    }
    source.neighbour = neighbourhood_generate(nn, peers, num_peers, -1);
    source.neigh_size = nn;
    source_set(source.neighbour, source.neigh_size, num_chunks);

    if (display_neighbourhood) {
      for (n = 0; n < num_peers; n++) {
        printf("Peer %d Neighbourhood: ", n);
        neighbourhood_print(peers[n].neighbour, peers[n].neigh_size);
        printf("\n");
      }
    }
    if (graphfile) {
      graph_dotprint(graphfile, &source, peers, n);
    }
    printf("Buf Size: %d\n", buf_size);
    ch = td_loop(peers, num_peers, num_chunks, ts);

    if (!per_chunk_delay_analysis(NULL, ch, num_chunks, num_peers)){
      printf("sub-optimal distribution!!!\n");
    }
    if (trace){
      per_chunk_delay_analysis(trace, ch, num_chunks, num_peers);
      per_node_delay_analysis(trace,  ch, num_chunks);
    }

    if (statsfile) {
      fprintf(statsfile, "Seed: %u\n", seed);
      per_chunk_delay_analysis(statsfile, ch, num_chunks, num_peers);
      per_node_delay_analysis(statsfile,  ch, num_chunks);
    }

  return 0;
}
