#!/bin/bash
#********************************************************************
# Copyright (C) 2008  University of Trento
#
# Author:    Csaba Kiraly
#  	     surname (at) disi.unitn.it
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#********************************************************************

# Usage: set paramters in this script (after the usage function), then run it with the command line option below
function usage () {
   echo "Usage: $PROGNAME -d results_dir -t testcase [-p filename_prefix] [-s src_dir] [-c]"
   echo "        -d dir: directory for result files"
   echo "        -t test: name of the test to execute"
   echo "        -p prefix: filename prefix of result files"
   echo "        -s dir: directory with sources, used also for recompile"
   echo "        -c: copy sources to results directory, and compile there"
   exit $1
}
# examples:
#
# ./test.sh -d results -t chunkdelay-cdf
# Starts recompiling the program in place, and puts results in the "./results" directory.
# It runs the "chunkdelay-cdf" test case, i.e. it genates a CDF for all given combination of chunk and peer scheduling algorithms
# a copy of ./test.sh (i.e. paramters) will be saved in the "results" folder
# a script "rerun.sh" will also be created in the "results" directory, keeping track of the command line options below
#
# ./test.sh -d results2/047-bw1-homogbw-deg3-numnodes -p 047-bw1-homogbw-deg3_ -t numnodes -c
# -c means that it will make a snapshot of the sources under "results2/047-bw1-homogbw-deg3-numnodes/src"
# these sources will be used for compilation which will also be done in the new folder, thus it will not interfere with other simulations or modifications to the source
# finally, it goes into background.

# set common parameters below. These will be later overridden by the testcase, if needed.
# a lower case "s" at the end of the parameter name means that the parameter is a list, i.e. the script will iterate through all the given values

CYCLEs=1                # don't change, it does not make sense with the current code: seed guarantees deterministic behaviour, independent of the cycle
NUM_SEEDS=10             # number of random graphs to generate for the test. Note that MAX values will be averaged amond runs with different seeds
SEEDs=`seq 1 1 $NUM_SEEDS`
NUM_ITERATIONS=3        # number of runs done (with each algorithm) on the same graph. MAX values will be maxed among these runs

NUM_PEERSs=100          # number of peers, including the source
DEGREEs='($NUM_PEERS)'   # neighborhood size. Set in to "NUM_PEERS-1" for a full mesh. To be improved when better graph generation algorithms are implemented
MEANBWs=1.0             # mean bandwidth. 1:critical regime; <1:overload regime >1:undeload regime
PEERSCLASS1s=1.0        # ratio of peers in the first class
MEANCLASS1s='(($MEANBW-(1.0-$PEERSCLASS1)*$MEANCLASS2)/$PEERSCLASS1)' # BW of peers in the first class. It can be calcualted based on MEANBW = PC1*MC1 + (1-PC1)*MC2
PEERSCLASS2s='(1-$PEERSCLASS1)'
MEANCLASS2s=0.5        

NUM_CHUNKSs=200         # number of chunks

CSCHEDs="LUc MinDLc RUc" # #RUsLUc RUc EUsLUc         # chunk scheduling alogorithms used
PSCHEDs="U_LNNC_p RUp MDp" #RUp U_LC_p U_LNNUC_p BAp BAp_w U_DOWNUP_p U_STRIPE_p            # peer scheduling alogorithms used
WIN_SIZEs="50" # "NUM_CHUNKS"  # trading window size. Set to "NUM_CHUNKS" to disable
PLAYOUT_DEADLINEs="-1"  # playout deadline: chunk having more delay are not needed and dropped. Set to "-1" to disable

#==================================================================
HEADINGLINE="#cycle\tseed\tpeers\tdegree\tpeercl1\tbw1\tpeercl2\tbw2\tchunks\tcscheduler\tpscheduler\twinsize\tTESTresults\tlossratio\tavgdelay\tmaxdelay"
LINEHEAD='$CYCLE\t$SEED\t$NUM_PEERS\t$DEGREE\t$PEERSCLASS1\t$MEANCLASS1\t$PEERSCLASS2\t$MEANCLASS2\t$NUM_CHUNKS\t$CSCHED\t$PSCHED\t$WIN_SIZE' #will be evaluated later using eval
RESULTSINDEX=13

#==================================================================
#==================================================================
function configure_testcase {
case $1 in 
  winsize)
    XLABEL="\"window size (in chunks)\""
    XSHORTNAME=winsize
    XINDEX=12
    PLOTTITLE='Effect of window size \n
     homogenous bandwith, set to $MEANBWs \n
     peers=$NUM_PEERSs; degree=$DEGREEs; chunks=$NUM_CHUNKSs; iterations=$NUM_ITERATIONS; seeds=$NUM_SEEDS; window=$WIN_SIZEs'    #will be evaluated later using eval
    WIN_SIZEs="5 10 15 20 30 40 50 70 100 190 195 200"
    ;;

  highbwratio)
    XLABEL="\"ratio of high BW nodes (inhomogeneous)\""
    XSHORTNAME=highbwratio
    XINDEX=5
    PLOTTITLE='Effect of inhomogeneity \n
     average BW fixed at $MEANBWs, low BW fixed at $MEANCLASS2s \n
     peers=$NUM_PEERSs; degree=$DEGREEs; chunks=$NUM_CHUNKSs; iterations=$NUM_ITERATIONS; seeds=$NUM_SEEDS; window=$WIN_SIZEs'    #will be evaluated later using eval
    MEANBWs=1.0
    PEERSCLASS1s=`seq 0.05 0.05 1`
    MEANCLASS2s=0.5        
    ;;

  highbw)
    XLABEL="\"high BW (note that the number of such nodes decreases, the average is always the same)\""
    XSHORTNAME=highbw
    XINDEX=6
    PLOTTITLE='Effect of inhomogeneity \n
     average BW fixed at $MEANBWs, low BW fixed at $MEANCLASS2s \n
     peers=$NUM_PEERSs; degree=$DEGREEs; chunks=$NUM_CHUNKSs; iterations=$NUM_ITERATIONS; seeds=$NUM_SEEDS; window=$WIN_SIZEs'    #will be evaluated later using eval
    MEANBWs=1.0
    PEERSCLASS1s=`seq 0.05 0.05 1`
    MEANCLASS2s=0.5        
    ;;

  bw)
    XLABEL="\"homogeneous BW\""
    XSHORTNAME=bw
    XINDEX=6
    PLOTTITLE='Effect of homogeneous bandwidth\n
     peers=$NUM_PEERSs; degree=$DEGREEs; chunks=$NUM_CHUNKSs; iterations=$NUM_ITERATIONS; seeds=$NUM_SEEDS; window=$WIN_SIZEs'    #will be evaluated later using eval
    MEANBWs=`seq 0.9 0.10 3`
    PEERSCLASS1s=1.0
    ;;

  numnodes)
    XLABEL="\"Number of peers\""
    XSHORTNAME=numnodes
    XINDEX=3
    PLOTTITLE='Number of peers\n
     homogenous bandwith, set to $MEANBWs \n
     peers=$NUM_PEERSs; degree=$DEGREEs; chunks=$NUM_CHUNKSs; iterations=$NUM_ITERATIONS; seeds=$NUM_SEEDS; window=$WIN_SIZEs'    #will be evaluated later using eval
    #NUM_PEERSs="2 3 5 7 10 15 20 30 50 70 100"
    NUM_PEERSs="200 300 500 700 1000 1500 2000 3000 5000 7500 10000"
    ;;

  degree)
    XLABEL="\"Neighborhood size\""
    XSHORTNAME=degree
    XINDEX=4
    PLOTTITLE='Neighborhood size\n
     homogenous bandwith, set to $MEANBWs \n
     peers=$NUM_PEERSs; degree=$DEGREEs; chunks=$NUM_CHUNKSs; iterations=$NUM_ITERATIONS; seeds=$NUM_SEEDS; window=$WIN_SIZEs'    #will be evaluated later using eval
    #DEGREEs="5 7 10 15 20 30 50 70 99"
    DEGREEs="10 13 14 15  50 100 250 500 750 1000 1500 2000 3000 5000 7000 10000"
    ;;

  numchunks)
    XLABEL="\"Stream length (number of chunks)\""
    XSHORTNAME=numchunks
    XINDEX=9
    PLOTTITLE='Stream length (number of chunks)\n
     homogenous bandwith, set to $MEANBWs \n
     peers=$NUM_PEERSs; degree=$DEGREEs; chunks=$NUM_CHUNKSs; iterations=$NUM_ITERATIONS; seeds=$NUM_SEEDS; window=$WIN_SIZEs'    #will be evaluated later using eval
    #NUM_CHUNKSs="5 20 50 100 200 500 1000 5000"
    NUM_CHUNKSs="50 200 500 1000 2000 5000 10000 15000 20000"
    ;;

  chunkdelay-cdf)
    XLABEL="\"chunk delay\""
    XSHORTNAME=delay
    PLOTTITLE='CDf of individual chunk delays\n
     homogenous bandwith, set to $MEANBWs \n
     peers=$NUM_PEERSs; degree=$DEGREEs; chunks=$NUM_CHUNKSs; iterations=$NUM_ITERATIONS; seeds=$NUM_SEEDS; window=$WIN_SIZEs'    #will be evaluated later using eval
    OUTPREFIX='$CSCHED-$PSCHED'
    STAT_CDF=1
    ;;
  *)
    echo "wrong testcase. Look in $0 for more details :)"
    exit 1
    ;;

esac
}


function set_schedulers {
  if [ $CSCHED = "LUc" ];
   then
    CS=latest_useful_chunk
  elif [ $CSCHED = "RUc" ];
   then
    CS=random_useful_chunk
  elif [ $CSCHED = "MinDLc" ];
   then
    CS=dl_sched
   fi

  if [ $PSCHED = "BAp_w" ];
   then
    echo "Unsupported Peer Scheduler BA";
    exit
  elif [ $PSCHED = "BAp" ];
   then
    echo "Unsupported Peer Scheduler BA";
    exit
  elif [ $PSCHED = "U_LNNUC_p" ];
   then
    echo "Unsupported Peer Scheduler U_LNNUC_p";
    exit
  elif [ $PSCHED = "U_LNNC_p" ];
    then
     PS=ch_earliest_free_peer
  elif [ $PSCHED = "MDp" ];
    then
     PS=ch_most_deprived_peer
  elif [ $PSCHED = "U_LC_p" ];
   then
    echo "Unsupported Peer Scheduler U_LC_p";
    exit
  elif [ $PSCHED = "U_DOWNUP_p" ];
   then
    echo "Unsupported Peer Scheduler U_DOWNUP_p";
    exit
  elif [ $PSCHED = "U_STRIPE_p" ];
   then
    echo "Unsupported Peer Scheduler U_STRIPE_p";
    exit
  elif [ $PSCHED = "RUp" ];
   then
     PS=ch_useful_peer_random
  fi
}

function recompile {
  cd $BUILDDIR
  make -f $SRCDIR/Makefile SRC=$SRCDIR clean
  make -f $SRCDIR/Makefile SRC=$SRCDIR CHUNK_SCHED=$CS PEER_SCHED=$PS
  cd -
}

function run {
  eval LINEHEAD2=\"$LINEHEAD\"

  if [[ $RESUME && `awk "/$LINEHEAD2/" $TESTOUT` ]] ; then
    echo "$LINEHEAD2 skipped by resume"
    return
  fi

  echo -e  $LINEHEAD2
  RESULTSLINE=$LINEHEAD2
  echo running: $SIMULATOR $@
  #RESULTSLINE+=`/usr/bin/time $SIMULATOR $@ | grep "TESTresults"`
  #/usr/bin/time -f "user %U sys %S mem %M" -o time.out $SIMULATOR $@ | tee run.out
  #RESULTSLINE+=`grep TESTresults run.out`
  #RESULTSLINE+=" "`cat time.out`
  $BUILDDIR/proc-time 2>&1 $SIMULATOR $@ | tee run.out
  RESULTSLINE+=`grep TESTresults run.out`
  RESULTSLINE+=" "`grep user run.out`
  RESULTSLINE+=" "`grep MAXSIZE run.out`
  RESULTSLINE+=" "`grep AVGSIZE run.out`

  echo -e $RESULTSLINE >>$TESTOUT
  if [[ $OUTPREFIX ]]; then
    eval OUTPREFIX2=\"$OUTPREFIX\"
    cp Perchunkinstance_cdf.out $OUTPREFIX2-Perchunkinstance_cdf.out
  fi
}


#parameters plotstr, title, outfile_prefix
function gnuplot_cdf {
cat >$3.gnuplot <<INPUT_END
set style data steps
set title "$2"
set xlabel $XLABEL
set ylabel $YLABEL

set terminal postscript eps color
set output "$3.eps"

$1

set terminal png size 1024,768
set output "$3.png"
replot
INPUT_END

gnuplot $3.gnuplot

}

#parameter: prefix
function plot_cdf {
  eval PLOTTITLE2=\"$PLOTTITLE\"
  PLOTSTR=""

  for CSCHED in $CSCHEDs ; do
  PLOTSTR2=""
  for PSCHED in $PSCHEDs ; do
    eval OUTPREFIX2=\"$OUTPREFIX\"
    TITLE=$OUTPREFIX2
    FILE="$OUTPREFIX2-Perchunkinstance_cdf.out"
    if [ "$PLOTSTR" ] ; then PLOTSTR+=$'\n replot' ; else PLOTSTR+='plot' ;  fi
    PLOTSTR="$PLOTSTR \"$FILE\" $USING title \"$TITLE\""
    if [ "$PLOTSTR2" ] ; then PLOTSTR2+=$'\n replot' ; else PLOTSTR2+='plot' ;  fi
    PLOTSTR2="$PLOTSTR2 \"$FILE\" $USING title \"$TITLE\""
  done;  
  gnuplot_cdf "$PLOTSTR2" "$PLOTTITLE2" $1-$CSCHED
  done;  

  gnuplot_cdf "$PLOTSTR" "$PLOTTITLE2" $1 

  for PSCHED in $PSCHEDs ; do
  PLOTSTR2=""
  for CSCHED in $CSCHEDs ; do
    eval OUTPREFIX2=\"$OUTPREFIX\"
    TITLE=$OUTPREFIX2
    FILE="$OUTPREFIX2-Perchunkinstance_cdf.out"
    if [ "$PLOTSTR2" ] ; then PLOTSTR2+=$'\n replot' ; else PLOTSTR2+='plot' ;  fi
    PLOTSTR2="$PLOTSTR2 \"$FILE\" $USING title \"$TITLE\""
  done;  
  gnuplot_cdf "$PLOTSTR2" "$PLOTTITLE2" $1-$PSCHED
  done;  
}

#parameter: prefix
function plot_one {
  PLOTSTR=""

  for CURVE1 in $CSCHEDs ; do
  for CURVE2 in $PSCHEDs ; do
    CASE="\t$CURVE1\t$CURVE2\t"
    TITLE="$CURVE1-$CURVE2"
    FILE="< awk '/$CASE/' $TESTOUT"
    if [ "$PLOTSTR" ] ; then PLOTSTR+=$'\n replot' ; else PLOTSTR+='plot' ;  fi
    PLOTSTR="$PLOTSTR \"$FILE\" $USING title \"$TITLE\""
  done;  
  done;  

  eval PLOTTITLE2=\"$PLOTTITLE\"

#prepare EPS
cat >$1.gnuplot <<INPUT_END
set style data yerrorlines
set title "$PLOTTITLE2"
set xlabel $XLABEL
set ylabel $YLABEL

$PLOTSTR

set terminal postscript eps color
set output "$1.eps"
replot

set terminal png size 1024,768
set output "$1.png"
replot
INPUT_END


}

function plotfiles {
if [[ $OUTPREFIX ]]; then
  YLABEL="\"CDF\""
  YSHORTNAME=cdf
  USING=""   #requires gnuplot stddev patch!
  plot_cdf ${PREFIX}${XSHORTNAME}-${YSHORTNAME}
else
  YLABEL="\"average delay [s]\""
  YSHORTNAME=avgdelay
  YINDEX=$(($RESULTSINDEX + 2))
#  USING="using $XINDEX:$YINDEX:$YINDEX:$YINDEX smooth unique"   #requires gnuplot stddev patch!
  USING="using $XINDEX:$YINDEX smooth unique"
  plot_one ${PREFIX}${XSHORTNAME}-${YSHORTNAME}

  YLABEL="\"max delay [s]\""
  YSHORTNAME=maxdelay
  YINDEX=$(($RESULTSINDEX + 3))
  USING="using $XINDEX:$YINDEX smooth unique"
  plot_one ${PREFIX}${XSHORTNAME}-${YSHORTNAME}

  YLABEL="\"chunk loss ratio\""
  YSHORTNAME=chunkloss
  YINDEX=$(($RESULTSINDEX + 1))
  USING="using $XINDEX:$YINDEX smooth unique"
  plot_one ${PREFIX}${XSHORTNAME}-${YSHORTNAME}
fi
}

function plot {
  for GNUPLOT_FILE in `ls *.gnuplot` ; do
	gnuplot $GNUPLOT_FILE
  done;
}

#===============================================================================


PROGNAME=$0
CMDLINE="$0 $*"

while getopts "S:d:p:cs:t:brh" opt; do
   case $opt in

   d )  PREFIXDIR=$OPTARG ;;
   p )  PREFIX=$OPTARG ;;
   s )  BUILDDIR=$OPTARG ;;
   c )  COPYSOURCE=1;;
   t )  TESTCASE=$OPTARG ;;
   b )  BACKGROUND=1;;
   S )  SRCDIR=$OPTARG;;
   r )  RESUME=1;;
   h )  usage 0 ;;
   \?)  usage 1 ;;
   esac
done

shift $(($OPTIND - 1))
if [[ $1 || ! $TESTCASE || ! $PREFIXDIR ]]; then
  usage 1
fi

[[ $BUILDDIR ]] || BUILDDIR=`pwd`

[[ $SRCDIR ]] || SRCDIR=`pwd`

SIMULATOR=$BUILDDIR/sssim
TESTOUT=${PREFIX}results.txt

mkdir -p $PREFIXDIR

#copy sources if needed
if [[ $COPYSOURCE ]]; then
  #git clone --depth 1 . "$PREFIXDIR/src" || exit 1
  SRCDIR2=`pwd`/"$PREFIXDIR/src"
  cd $SRCDIR
  git clone . "$SRCDIR2" || exit 1
  git diff | patch -d "$SRCDIR2" -p1 || exit 1
  cd -
  cd $PREFIXDIR
  exec $SRCDIR2/`basename $0` -d . -p $PREFIX -t $TESTCASE -S $SRCDIR2 -s $SRCDIR2 -b -r &
  exit 0
else 
  cd $PREFIXDIR
fi

#save command line parameters
echo $CMDLINE >${PREFIX}rerun.sh
chmod u+x ${PREFIX}rerun.sh

#save code with parameters for later reference
#cp $SRCDIR/$0 ${PREFIX}`basename $0`

# redirect stdout(1) and stderr(2) to logfile and stdout:
if [[ $BACKGROUND ]]; then
  exec 2>${PREFIX}.err 1>${PREFIX}.out
else
  exec 2> >(tee ${PREFIX}.err) 1> >(tee ${PREFIX}.out)
fi

#set parameters for the given test case
configure_testcase $TESTCASE

#create results file
if [[ ! $RESUME || ! -f $TESTOUT ]]; then
  echo -e $HEADINGLINE >$TESTOUT
fi

#create gnuplot files
plotfiles

#start the big cycle
for CYCLE in $CYCLEs ; do
for SEED in $SEEDs ; do
for MEANBW in $MEANBWs ; do
for PEERSCLASS1 in $PEERSCLASS1s ; do
  PEERSCLASS1=$(eval echo \"scale=6 \; $PEERSCLASS1\" | bc)
for MEANCLASS2 in $MEANCLASS2s ; do        
  MEANCLASS2=$(eval echo \"scale=6 \; $MEANCLASS2\" | bc)
for MEANCLASS1 in $MEANCLASS1s ; do
  MEANCLASS1=$(eval echo \"scale=6 \; $MEANCLASS1\" | bc)
for PEERSCLASS2 in $PEERSCLASS2s ; do
  PEERSCLASS2=$(eval echo \"scale=6 \; $PEERSCLASS2\" | bc)
for WIN_SIZE in $WIN_SIZEs ; do
for PLAYOUT_DEADLINE in $PLAYOUT_DEADLINEs ; do
for CSCHED in $CSCHEDs ; do
for PSCHED in $PSCHEDs ; do
  date
  set_schedulers 
  recompile $BUILDDIR
for NUM_PEERS in $NUM_PEERSs ; do
for DEGREE in $DEGREEs ; do
  DEGREE=$(eval echo \"scale=6 \; $DEGREE\" | bc)
for NUM_CHUNKS in $NUM_CHUNKSs ; do
  run -s $SEED -n $NUM_PEERS -c $NUM_CHUNKS -o $DEGREE
done;
done;
done;
done;
done; #SCHED
plot
done;
done;
done;
done;
done;
done;
done;
done;
done;
